#include <openssl/bn.h>
#include <stdio.h>
#include <sys/time.h>

#define NUMBITS 2000
#define SLICESZ  600

/*
 *  This isn't the primeSearch you'd write for a sequential computation..
 *
 *  primeSearch(BIGNUM *t, int sliceSize)
 *    Searches for primes in the range t to t+sliceOffset
 *      of the form t + 6 n
 *    Assumes that t mod 6 is either 1 or 5
 *    Assumes that sliceSize mod 6 is 0
 *    If a prime is found, 
 *      returns 1 and sets t to the found prime, 
 *    If a prime is *not* found, 
 *      returns 1 and sets t to t+sliceOffset
 */
int primeSearch(BIGNUM *t, int sliceSize) {
  int i ;
  for (i=0; i<sliceSize; i+=6) {
    if (BN_is_prime(t, BN_prime_checks, NULL, NULL, NULL))
      return 1 ;
    else
      (void) BN_add_word(t, (BN_ULONG) 6) ;
  }
  return 0 ;
}

/* return n mod 6 as an integer */
unsigned int mod6(BIGNUM *n) {
  BIGNUM *vi, *r ;
  BN_CTX *tmpStuff ;
  int rInt ;
  tmpStuff = BN_CTX_new() ;
  r  = BN_new () ;
  vi = BN_new() ;
  (void) BN_set_word(vi, (unsigned long)6) ;
  (void) BN_div((BIGNUM *)NULL, r, n, vi, tmpStuff) ;
  rInt = (unsigned int) BN_get_word(r) ;
  BN_CTX_free(tmpStuff) ;
  BN_free(vi) ;
  BN_free(r) ;
  return(rInt) ;
}

int vectP1[6] = { 1, 0, 3, 2, 1, 0 } ;
int vectP2[6] = { 5, 4, 5, 4, 3, 2 } ;

int main(int argc, char *argv[]) {
  struct timeval startTime, endTime ;
  int runMSec, stillLooking, sliceStart, m6, offP1, offP2 ;
  BIGNUM *p ;

  p = BN_new() ;
  (void) BN_rand(p, NUMBITS, 1, 1) ;

  m6 = mod6(p) ;
  offP1 = vectP1[m6] ;
  offP2 = vectP2[m6] ;

  gettimeofday(&startTime, NULL) ;
  while (1) {
    /* 
     * Duplicating p may seem pointless (or -- pointerful)
     * However, it the algorithm was "parallelized"
     * we could then run primeSearch on a second processor
     *
     */
    BIGNUM *copyP1, *copyP2 ;

    /* one-half the available numbers */
    copyP1 = BN_dup(p) ;
    (void) BN_add_word(copyP1, (BN_ULONG) offP1) ;
    if (primeSearch(copyP1, SLICESZ)) {
      (void) BN_copy(p, copyP1) ;
      break ;            /* breaks are ugly */
    } else 
      offP1 += SLICESZ ;

    /* the other half */
    copyP2 = BN_dup(p) ;
    (void) BN_add_word(copyP2, (BN_ULONG) offP2) ;
    if (primeSearch(copyP2, SLICESZ)) {
      (void) BN_copy(p, copyP2) ;
      break ;            /* breaks are ugly */
    } else 
      offP2 += SLICESZ ;
  }
  gettimeofday(&endTime, NULL) ;

  (void) BN_print_fp(stdout, p) ;
  runMSec = (endTime.tv_sec - startTime.tv_sec) * 1000 +
    (endTime.tv_usec - startTime.tv_usec + 500) / 1000 ;
  fprintf(stdout, "\n  Time = %d msec\n", runMSec) ;
}
